import itertools
import json
import logging
import time
from typing import Text, List, Dict, Any, Optional, Tuple, Set

from sqlalchemy import or_, and_, true, false

from rasa.core.constants import UTTER_PREFIX
from rasax.community import config
from rasax.community.database.data import Template
from rasax.community.database.service import DbService
from rasax.community.utils import (
    get_columns_from_fields,
    get_query_selectors,
    query_result_to_dict,
    QueryResult,
)

logger = logging.getLogger(__name__)


class NlgService(DbService):
    def save_template(
        self,
        template: Dict[Text, Any],
        username: Optional[Text] = None,
        domain_id: Optional[Text] = None,
        project_id: Text = config.project_name,
    ) -> Dict[Text, Any]:
        """Save response.

        Args:
            template: Response object to save.
            username: Username performing the save operation.
            domain_id: Domain associated with the response.
            project_id: Project ID associated with the response.

        Raises:
            `ValueError` if any of:
            1. no domain ID is specified
            2. the response name does not start with `rasa.core.constants.UTTER_PREFIX`
            3. a response with the same text already exists

        Returns:
            The saved response object.
        """
        if domain_id is None:
            raise ValueError("Template could not be saved since domain ID is `None`.")

        response_name = Template.get_stripped_value(template, "template")

        if not response_name or not response_name.startswith(UTTER_PREFIX):
            raise AttributeError(
                f"Failed to save response. Response '{response_name}' does "
                f"not begin with '{UTTER_PREFIX}' prefix."
            )

        template["template"] = response_name
        template["text"] = Template.get_stripped_value(template, "text")

        if self.get_template(project_id, template):
            raise ValueError(
                f"Another template with same pair of (template, text) "
                f"(({response_name},{template['text']})) already exists for this "
                f"project and domain."
            )

        if not username:
            username = config.default_username

        new_template = Template(
            template=template["template"],
            text=template["text"],
            content=json.dumps(template),
            annotated_at=time.time(),
            annotator_id=username,
            project_id=project_id,
        )

        if domain_id:
            new_template.domain_id = domain_id

        self.add(new_template)

        # flush so ID becomes available
        self.flush()

        return new_template.as_dict()

    def fetch_templates(
        self,
        text_query: Optional[Text] = None,
        template_query: Optional[Text] = None,
        fields_query: List[Tuple[Text, bool]] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        sort_by_template_name: bool = False,
        intersect_filters: bool = False,
    ) -> QueryResult:
        """Returns a list of responses. Each response includes its template name
        as a property under "template".

        An example of a response item could be:
        ```
        {
            "text": "Hey! How are you?",
            "template": "utter_greet",
            "id": 6,
            "annotator_id": "me",
            "annotated_at": 1567780833.8198001,
            "project_id": "default",
            "edited_since_last_training": false
        }
        ```

        Args:
            text_query: Filter results by response text (case insensitive).
            template_query: Filter results by template name.
            fields_query: Fields to include per item returned.
            limit: Maximum number of responses to be retrieved.
            offset: Excludes the first ``offset`` responses from the result.
            sort_by_template_name: If ``True``, sort responses by their
                "template" property.
            intersect_filters: If ``True``, join ``text_query`` and
                ``template_query`` conditions with an AND instead of an OR.

        Returns:
            List of responses with total number of responses found.
        """

        templates_to_query = template_query.split(",") if template_query else []
        columns = get_columns_from_fields(fields_query)

        if not templates_to_query and not text_query:
            # Case 1: No conditions were specified - base query is TRUE (all results)
            query = true()
        else:
            if intersect_filters:
                # Case 2: One or both conditions were specified, and caller wants
                #         to join them with AND.
                query = true()
            else:
                # Case 3: One or both conditions were specified, and caller wants
                #         to join them with OR.
                query = false()

        query_joiner = and_ if intersect_filters else or_

        if text_query:
            query = query_joiner(query, Template.text.ilike(f"%{text_query}%"))

        if template_query:
            query = query_joiner(query, Template.template.in_(templates_to_query))

        templates = self.query(*get_query_selectors(Template, columns)).filter(query)

        if sort_by_template_name:
            templates = templates.order_by(Template.template.asc())

        total_number_of_results = templates.count()

        templates = templates.offset(offset).limit(limit).all()

        if columns:
            results = [query_result_to_dict(r, fields_query) for r in templates]
        else:
            results = [t.as_dict() for t in templates]

        return QueryResult(results, total_number_of_results)

    def get_grouped_responses(
        self, text_query: Optional[Text] = None, template_query: Optional[Text] = None
    ) -> QueryResult:
        """Returns responses grouped by their template name.

        Args:
            text_query: Filter responses by response text (case insensitive).
            template_query: Filter template groups by template name.

        Returns:
            `QueryResult` containing grouped responses and total number of responses
            across all groups.
        """
        # Sort since `groupby` only groups consecutive entries
        templates = self.fetch_templates(
            text_query=text_query,
            template_query=template_query,
            sort_by_template_name=True,
            intersect_filters=True,
        ).result

        grouped_templates = itertools.groupby(
            templates, key=lambda each: each["template"]
        )

        result = [{"template": k, "responses": list(g)} for k, g in grouped_templates]
        count = sum(len(item["responses"]) for item in result)

        return QueryResult(result, count)

    def fetch_all_template_names(self) -> Set[Text]:
        """Fetch a list of all template names in db."""

        return {t["template"] for t in self.fetch_templates()[0]}

    def delete_template(self, _id: int) -> bool:
        delete_result = self.query(Template).filter(Template.id == _id).delete()

        return delete_result

    def delete_all_templates(self) -> None:
        self.query(Template).delete()

    def update_template(
        self, _id: int, template: Dict[Text, Any], username: Text
    ) -> Optional[Template]:
        old_template = self.query(Template).filter(Template.id == _id).first()
        if not old_template:
            raise KeyError(f"Could not find existing response with ID '{_id}'.")

        response_name = Template.get_stripped_value(template, "template")

        if not response_name or not response_name.startswith(UTTER_PREFIX):
            raise AttributeError(
                f"Failed to update response. Response '{response_name}' does "
                f"not begin with '{UTTER_PREFIX}' prefix."
            )

        template["template"] = response_name
        template["text"] = Template.get_stripped_value(template, "text")

        if self.get_template(old_template.project_id, template):
            raise ValueError(
                "Template could not be saved since another one with same pair of "
                "(template, text) exists."
            )

        old_template.template = template["template"]
        old_template.text = template["text"]
        old_template.annotated_at = time.time()
        old_template.content = json.dumps(template)
        old_template.annotator_id = username
        old_template.edited_since_last_training = True

        # because it's the same as the updated template
        return old_template

    def replace_templates(
        self,
        new_templates: List[Dict[Text, Any]],
        username: Optional[Text] = None,
        domain_id: Optional[Text] = None,
    ) -> int:
        """Deletes all responses and adds responses from template_list.

        Returns the number of inserted templates.
        """
        self.delete_all_templates()
        insertions = 0
        for template in new_templates:
            try:
                inserted = self.save_template(template, username, domain_id=domain_id)
                if inserted:
                    insertions += 1
            except ValueError:
                pass

        return insertions

    def get_template(
        self, project_id: Text, template: Dict[Text, Any]
    ) -> Optional[Template]:
        """Return a template that has the specified `project_id` and `template`.

        Args:
            project_id: Project ID.
            template: Template.
        """
        template_content = json.dumps(template)

        return (
            self.query(Template)
            .filter(
                and_(
                    Template.project_id == project_id,
                    Template.content.like(f"%{template_content}%"),
                )
            )
            .first()
        )

    def mark_templates_as_used(
        self, training_start_time: float, project_id: Text = config.project_name
    ) -> None:
        """Unset the `edited_since_last_training_flag` for templates which were included
        in the last training.

        Args:
            training_start_time: annotation time until templates should be marked as
                used in training.
            project_id: Project which was trained.

        """
        templates_which_were_used_in_training = (
            self.query(Template)
            .filter(
                and_(
                    Template.annotated_at <= training_start_time,
                    Template.edited_since_last_training,
                    Template.project_id == project_id,
                )
            )
            .all()
        )
        for template in templates_which_were_used_in_training:
            template.edited_since_last_training = False

    def rename_templates(
        self, old_template_name: Text, template: Dict[Text, Text], annotator: Text
    ) -> None:
        """
        Bulk renames a template name.

        Args:
            old_template_name: The name of the template which should be renamed.
            template: The object containing the new template values.
            annotator: The name of the user who is doing the rename.
        """
        templates = (
            self.query(Template).filter(Template.template == old_template_name).all()
        )
        new_template_name = template["name"]
        for template in templates:
            template.template = new_template_name
            template.annotated_at = time.time()
            template.annotator_id = annotator
